#ifndef _AN_HPP_
#define _AN_HPP_

/* A class to handle numbers in an algebraic extension of Q */

#include "u.hpp"
#include "inter.hpp"

using namespace std;

struct poly_t;
void div(const poly_t &a,const poly_t &b,poly_t &q,poly_t &r);
ostream &operator<<(ostream &o,const poly_t &p);

struct poly_t { //polynom over rationnals
  map<int,rat_t> x;

  poly_t(rat_t i=0,int d=0) {
    x[d]=i;
  }

  bool isnull() const {
    for(auto &it:x)
      if(it.second!=0) return 0;
    return 1;
  }

  int deg() const {
    int r=-1;
    for(auto &it:x) {
      if(it.second==0) continue;
      if(it.first>r) r=it.first;
    }
    return r;
  }

  rat_t &operator[](int d) {
    return x[d];
  }

  rat_t operator[](int d) const {
    auto it=x.find(d);
    if(it==x.end()) return 0;
    return it->second;
  }

  void simpl() {
    map<int,rat_t > x2;
    for(auto &it:x)
      if(it.second)
	x2[it.first]=it.second;
    x.swap(x2);
  }

  poly_t der() const {
    poly_t r;
    for(auto &it:x)
      if(it.second!=0 && it.first>0) {
	r.x[it.first-1]+=it.first*it.second;
      }
    return r;
  }

  long double eval(long double a) const {
    long double r=0;
    for(auto &it:x)
      r+=it.second.get_d()*powl(a,it.first);
    return r;
  }

  static rat_t pow(const rat_t &a,int i) {
    rat_t x=1;
    for(int u=0;u<i;u++)
      x*=a;
    return x;
  }

  rat_t eval(const rat_t &a) const {
    rat_t r=0;
    for(auto &it:x)
      r+=it.second*pow(a,it.first);
    return r;
  }

  inter_t bnd(const rat_t &x,const rat_t &l) const { // returns a bound of P(a+x) |x|<=l
    int d=deg();
    if(d<0) return inter_t(0);
    if(d==0) return inter_t(this->x.at(0));
    poly_t de=der();
    inter_t in=de.bnd(x,l);
    rat_t mb=in.b;
    if(mb<0) mb=-mb;
    rat_t ma=in.a;
    if(ma<0) ma=-ma;
    if(ma<mb) ma=mb;
    rat_t y=eval(x);
    return inter_t(y-l*ma,y+l*ma);
  }

  poly_t operator+(const poly_t &b) const {
    poly_t r=*this;
    for(auto &it:b.x)
      r.x[it.first]+=it.second;
    r.simpl();
    return r;
  }

  const poly_t &operator+=(const poly_t &b) {
    for(auto &it:b.x)
      x[it.first]+=it.second;
    simpl();
    return *this;
  }

  const poly_t &operator-=(const poly_t &b) {
    for(auto &it:b.x)
      x[it.first]-=it.second;
    simpl();
    return *this;
  }

  poly_t operator-(const poly_t &b) const {
    poly_t r=*this;
    for(auto &it:b.x)
      r.x[it.first]-=it.second;
    r.simpl();
    return r;
  }

  poly_t operator-() const {
    poly_t r;
    for(auto &it:x)
      r.x[it.first]=-it.second;
    return r;
  }

  poly_t operator*(const poly_t &b) const {
    poly_t r;
    for(auto &it:x)
      for(auto &jt:b.x)
	r.x[it.first+jt.first]+=it.second*jt.second;
    r.simpl();
    return r;
  }

  void dicho(rat_t &a,rat_t &b) const {
    rat_t ea=eval(a);
    rat_t eb=eval(b);
    assert(ea*eb<0);
    rat_t c=(a+b)/2;
    rat_t ec=eval(c);
    if(ea*ec<0) b=c;
    else a=c;
  }

  void dicho(inter_t &i,const rat_t &b) const {
    while(i.len()>b)
      dicho(i.a,i.b);
  }

  bool operator==(const poly_t &b) const {
    poly_t r=*this-b;
    return r.isnull();
  }
};

ostream &operator<<(ostream &o,const poly_t &p) {
  if(p.isnull()) {o<<"0";return o;}
  int f=1;
  for(auto &it:p.x) {
    if(it.second==0) continue;
    if(!f && it.second>0)
      o<<"+";
    f=0;
    if(it.first==0)
      o<<it.second;
    else    if(it.first==1)
      o<<it.second<<"*x";
    else
      o<<it.second<<"*x^"<<it.first;
    o<<" ";
  }
  return o;
}

void div(const poly_t &a,const poly_t &b,poly_t &q,poly_t &r)
{
  int da=a.deg();
  int db=b.deg();
  if(da<db) {
    r=a;
    q=poly_t(0);
    return;
  }
  auto xa=a[da];
  auto xb=b[db];
  auto aa=a-poly_t(xa/xb,da-db)*b;
  assert(aa.deg()<da);
  div(aa,b,q,r);
  q=q+poly_t(xa/xb,da-db);
}

struct root_t { // a polynom with a interval for one root of the polynom
  poly_t p;
  inter_t z;

  root_t(poly_t &p_,const rat_t &a=0,const rat_t &b=1):p(p_),z(a,b) {  }
  root_t(poly_t &p_,const inter_t &i):p(p_),z(i) {  }
  root_t() {}

  int deg() const { return p.deg();}
  bool operator==(const root_t &b) const { return p==b.p && z==b.z;}

  void dicho(rat_t &a,rat_t &b) const{
    p.dicho(a,b);
  }

  void getbound(rat_t &a,rat_t &b,const rat_t &r) const{
    a=z.a;
    b=z.b;
    while(b-a>r)
      dicho(a,b);
  }

  long double getd(int pr=10) const {
    rat_t a,b;
    rat_t r=1;
    for(int i=0;i<pr;i++)
      r/=10;
    getbound(a,b,r);
    return a.get_d();
  }
};

struct an_t;

ostream &operator<<(ostream &o,const an_t &p);

struct an_t { //a number on an algebraic extension of Q
  root_t modul;

  poly_t p;

  an_t(int i=0) {
    p=poly_t(i);
  }
  an_t(const poly_t &x) {
    p=x;
  }
  an_t(const poly_t &x,const root_t &mo) {
    modul=mo;
    p=x;
    simplify();
  }

  void simplify() {
    if(modul.deg()<0) return;
    if(p.deg()>=modul.p.deg()) {
      poly_t foo,r;
      div(p,modul.p,foo,r);
      p=r;
    }
  }

  void setmodul(const root_t &a) {
    if(modul.deg()<0) {
      if(a.deg()>=0) modul=a;
      return;
    }
    if(a.deg()>=0) assert(modul==a);
  }

  an_t operator+(const an_t &b) const {
    an_t r;
    r.setmodul(modul);
    r.setmodul(b.modul);
    r.p=p+b.p;
    r.simplify();
    return r;
  }
  const an_t &operator+=(const an_t &b) {
    p+=b.p;
    setmodul(b.modul);
    simplify();
    return *this;
  }
  const an_t &operator-=(const an_t &b) {
    p-=b.p;
    setmodul(b.modul);
    simplify();
    return *this;
  }

  an_t operator-(const an_t &b) const {
    an_t r;
    r.setmodul(modul);
    r.setmodul(b.modul);
    r.p=p-b.p;
    r.simplify();
    return r;
  }

  an_t operator-() const {
    an_t r;
    r.p=-p;
    r.modul=modul;
    return r;
  }

  an_t operator*(const an_t &b) const {
    auto a=(p*b.p);
    an_t r;
    r.setmodul(modul);
    r.setmodul(b.modul);
    r.p=a;
    r.simplify();
    return r;
  }

  an_t &operator*=(const an_t &b) {
    *this=*this*b;
    return *this;
  }

  an_t &operator/=(const an_t &b) {
    *this=*this/b;
    return *this;
  }

  bool isnull() const {
    if(p.deg()<=0) {
      return 0==(p[0]);
    }
    assert(modul.deg()>=0);
    assert(p.deg()<modul.p.deg());
    return p.isnull();
  }

  an_t inv() const {
    int d=p.deg();
    assert(d>=0); //if(deg<0) we have a division by zero
    if(d==0) { // he have a rational number
      an_t a;
      a.p[0]=1/p[0];
      return a;
    }
    assert(modul.deg()>=0);

    //BEZOUT
    poly_t u[3];
    u[0]=poly_t(0);
    u[1]=poly_t(1);
    poly_t w[3];
    w[0]=modul.p;
    w[1]=p;
    while(!w[1].isnull()) {
      poly_t q,r;
      div(w[0],w[1],q,r);
      w[2]=w[0]-q*w[1];
      u[2]=u[0]-q*u[1];
      u[0]=u[1];u[1]=u[2];
      w[0]=w[1];w[1]=w[2];
    }
    assert(w[0].deg()==0);
    poly_t pp=u[0]*poly_t(1/w[0][0]);
    an_t a(pp,modul);
    assert((a*(*this)-poly_t(1)).isnull());
    return a;
  }

  an_t operator/(const an_t &b) const {
    return *this*b.inv();
  }

  bool ispositive() const {
    int d=p.deg();
    if(d<=0) return p[0]>0;
    rat_t z0=modul.z.a;
    rat_t z1=modul.z.b;
    while(1) {
      inter_t in=p.bnd((z0+z1)/2,(z1-z0)/2);
      if(in.a>0) return 1;
      if(in.b<0) return 0;
      modul.dicho(z0,z1);
    }
  }

  bool operator<(const an_t &b) const {
    an_t r=b-*this;
    return r.ispositive();
  }

  bool operator<=(const an_t &b) const {
    an_t r=b-*this;
    if(r.isnull()) return 1;
    return r.ispositive();
  }

  bool operator==(const an_t &b) const {
    an_t r=b-*this;
    return r.isnull();
  }

  bool operator!=(const an_t &b) const {
    an_t r=b-*this;
    return !(r.isnull());
  }

  bool operator>=(const an_t &b) const {
    an_t r=*this-b;
    if(r.isnull()) return 1;
    return r.ispositive();
  }

  bool operator>(const an_t &b) const {
    an_t r=*this-b;
    return r.ispositive();
  }
};

bool isnull(const an_t &a) { return a.isnull();}

long double getd(const an_t &p)
{
  if(p.modul.deg()<0) {
    assert(p.p.deg()<=0);
    return getd(p.p[0]);
  }
  return p.p.eval(p.modul.getd());
}

int anprintpoly=1;

ostream &operator<<(ostream &o,const an_t &p) {
  if(p.p.deg()<=0 || p.modul.deg()<0) o<<p.p;
  else {
    cout<<p.p.eval(p.modul.getd());
    if(anprintpoly) o<<" ("<<p.p<<")";
    if(anprintpoly==2) o<<" ["<<p.modul.p<<" "<<p.modul.z.a<<".."<<p.modul.z.b<<"]";
  }
  return o;
}

#endif
