#ifndef __ASS_HPP__
#define __ASS_HPP__

/* ass__t<F> is a class of an affine subspace of F^n (where F will be rationals numbers, or algebraic numbers)
   The subspace S is represented by a point 'o' in S, and a basis 'v' of size 'dim'
   Main functions :
   - add(vector,q) : cut the affine subspace by the hyperplane x*vector=q
   - comp(vector,q) : return true iff the affine subspace is in the hyperplane x*vector=q (and in this case, add(vector,q) has no effect)
   - norm() : put in a normal form

   ass_t = ass__t<rat_t>;
*/

#include "u.hpp"

/* input and output of rationals */

string linq(const rat_t &a)
{
  int n=a.get_num().get_si();
  int d=a.get_den().get_si();
  assert(a.get_num()==n && a.get_den()==d); //check for overflows
  char bf[100];
  if(d==1)
    snprintf(bf,99,"%d",n);
  else
    snprintf(bf,99,"%d/%d",n,d);
  return bf;
}

rat_t readq(const char *x)
{
  int a,b=1;
  int r=sscanf(x,"%d/%d",&a,&b);
  assert(r>=1);
  return rat_t(a)/b;
}

void getnext(const char *&x) {
  while(x[0]!=' ') {
    assert(x[0]!=0);
    x++;
  }
  while(x[0]==' ')
    x++;
  assert(x[0]!=0);
}

void read(const char *&x,vector<rat_t> &v,int n)
{
  for(int i=0;i<n;i++) {
    v[i]=readq(x);
    getnext(x);
  }
}

template <class q_t>
struct ass__t {
  vector<q_t> o;
  vector<vector<q_t> > v;
  int n,dim;

  ass__t (int n_) {
    clear(n_,n_);
  }

  template<class q2_t>
  ass__t(const ass__t<q2_t> &obj) {
    n=obj.n;
    dim=obj.dim;
    o.resize(obj.o.size());
    v.resize(obj.v.size());
    for(int j=0;j<dim;j++)
      v[j].resize(obj.v[j].size());
    for(int i=0;i<5;i++) {
      o[i]=q_t(obj.o[i]);
      for(int j=0;j<dim;j++)
	v[j][i]=q_t(obj.v[j][i]);
    }
  }

  void clear(int n_,int d_) {
    n=n_;
    dim=d_;
    o.resize(n);
    for(int i=0;i<n;i++)
      o[i]=q_t(0);
    v.resize(d_);
    for(int i=0;i<d_;i++) {
      v[i].resize(n);
      for(int j=0;j<n;j++)
	v[i][j]=q_t(0);
      v[i][i]=q_t(1);
    }
  }

  bool operator==(const ass__t &b) const {
    if(*this<b) return false;
    if(b<*this) return false;
    return true;
  }

  bool operator<(const ass__t &b) const {
    if(dim<b.dim) return 1;
    if(dim>b.dim) return 0;
    if(dim==-1) return 0;
    for(int i=0;i<n;i++) {
      if(o[i]<b.o[i]) return 1;
      if(o[i]>b.o[i]) return 0;
    }
    for(int j=0;j<dim;j++) {
      for(int i=0;i<n;i++) {
	if(v[j][i]<b.v[j][i]) return 1;
	if(v[j][i]>b.v[j][i]) return 0;
      }
    }
    return 0;
  }

  vector<q_t> proj(const vector<q_t> &it) const { //gives the R^n point from the coordinates in the basis
    vector<q_t> r(n);
    for(int i=0;i<n;i++) {
      r[i]=o[i];
      for(int j=0;j<dim;j++)
	r[i]+=it[j]*v[j][i];
    }
    return r;
  }

  bool compo(const q_t *t,q_t q) const {
    q_t r=0;
    for(int i=0;i<n;i++)
      r+=o[i]*t[i];
    return isnull(r-q);
  }

  bool compv(int u,const q_t *t) const {
    q_t r=0;
    for(int i=0;i<n;i++)
      r+=t[i]*v[u][i];
    return isnull(r);
  }

  bool comp(const q_t *t,q_t q) const {
    if(dim<0) return 0;
    if(!compo(t,q)) return 0;
    for(int i=0;i<dim;i++)
      if(!compv(i,t)) return 0;
    return 1;
  }

  bool add(const q_t *t,q_t q) {
    if(dim==-1) return 0;
    if(comp(t,q)) return 0;
    recomp(t,q);
    return 1;
  }

  bool add(const vector<int> &tt,q_t q) {
    if(dim==-1) return 0;
    q_t t[n];
    for(int i=0;i<n;i++)
      t[i]=tt[i];
    if(comp(t,q)) return 0;
    recomp(t,q);
    return 1;
  }

  bool add(const vector<q_t> &tt,q_t q) {
    if(dim==-1) return 0;
    q_t t[n];
    for(int i=0;i<n;i++)
      t[i]=tt[i];
    if(comp(t,q)) return 0;
    recomp(t,q);
    return 1;
  }

  bool comp(const vector<int> &tt,q_t q) const {
    q_t t[n];
    for(int i=0;i<n;i++)
      t[i]=tt[i];
    return comp(t,q);
  }

  bool comp(const vi_t &tt,q_t q) const {
    q_t t[n];
    for(int i=0;i<n;i++)
      t[i]=tt[i];
    return comp(t,q);
  }

  bool vcomp(const vvi_t &tt,q_t q) const {
    for(auto &it:tt)
      if(comp(it,q)==false) return false;
    return true;
  }

  bool add(const vi_t &tt,q_t q) {
    if(dim==-1) return 0;
    q_t t[n];
    for(int i=0;i<n;i++)
      t[i]=tt[i];
    if(comp(t,q)) return 0;
    recomp(t,q);
    return 1;
  }

  void recomp(const q_t *t,q_t q) {
    int u=0;
    q_t rv=q_t(0);
    for(;u<dim;u++) {
      rv=q_t(0);
      for(int i=0;i<n;i++)
	rv+=t[i]*v[u][i];
      if(!isnull(rv)) break;
    }
    if(u==dim) {
      dim=-1;
      return;
    }
    {
      q_t r=0;
      for(int i=0;i<n;i++)
	r+=o[i]*t[i];
      q_t a=(q-r)/rv;
      for(int i=0;i<n;i++)
	o[i]+=a*v[u][i];
    }

    for(int w=0;w<dim;w++) {
      if(w==u) continue;
      q_t r=0;
      for(int i=0;i<n;i++)
	r+=t[i]*v[w][i];
      q_t a=-r/rv;
      for(int i=0;i<n;i++)
	v[w][i]+=a*v[u][i];
    }

    for(int i=0;i<n;i++)
      v[u][i]=v[dim-1][i];
    dim--;
  }

  void swapv(int i,int j)
  {
    assert(i!=j && i>=0 && j>=0 && i<dim && j<dim);
    for(int u=0;u<n;u++) {
      auto tmp=v[i][u];
      v[i][u]=v[j][u];
      v[j][u]=tmp;
    }
  }

  void norm() {
    if(dim<=0) return;
    int k=0;
    for(int u=0;u<n;u++) {
      for(int j=k;j<dim;j++) {
	if(!isnull(v[j][u])) {
	  if(j!=k) {
	    swapv(j,k);
	    j=k;
	  }
	  auto a=v[k][u];
	  for(int z=0;z<n;z++)
	    v[k][z]/=a;
	  assert(isnull(v[k][u]-1));
	  auto b=o[u];
	  for(int z=0;z<n;z++)
	    o[z]-=b*v[k][z];
	  assert(isnull(o[u]));
	  for(int l=0;l<dim;l++)
	    if(l!=k) {
	      b=v[l][u];
	      for(int z=0;z<n;z++)
		v[l][z]-=b*v[k][z];
	      assert(isnull(v[l][u]));
	    }
	  k++;
	  break;
	}
      }
      if(k==dim) break;
    }
  }

  void suma(vector<q_t> &v) {
    auto v2=v;
    q_t a=0;
    for(int i=0;i<n;i++) {
      v[i]=a;
      a+=q_t(1)-v2[(i+1)%n];
    }
    assert(a==2);
  }

  void sumb(vector<q_t> &v) {
    auto v2=v;
    q_t a=0;
    for(int i=0;i<n;i++) {
      v[i]=a;
      a+=-v2[(i+1)%n];
    }
    assert(a==0);
  }

  void sum() { //corresponds to the s() operation in the paper, Section 4
    if(dim<0) return;
    suma(o);
    for(int i=0;i<dim;i++)
      sumb(v[i]);
  }

  static vector<q_t> perm(const vector<q_t> &v,int *p)
  {
    vector<q_t> r(5);
    for(int i=0;i<5;i++)
      r[i]=v[p[i]];
    return r;
  }

  void perm(int *p)
  {
    assert(n==5);
    if(dim<0) return;
    o=perm(o,p);
    for(int i=0;i<dim;i++)
      v[i]=perm(v[i],p);
  }

  /* input & output */

  static string lin(const vector<q_t> &v,int n) {
    string r;
    for(int i=0;i<n;i++)
      r+=linq(v[i])+" ";
    return r;
  }

  string lin() const {
    string r;
    char bf[100];
    snprintf(bf,99,"%d %d  ",n,dim);
    r=string(bf)+lin(o,n)+"  ";
    for(int i=0;i<dim;i++)
      r+=lin(v[i],n)+"  ";
    return r;
  }

  void read(const char *x)
  {
    int nn,dd;
    int r=sscanf(x,"%d %d",&nn,&dd);
    assert(r==2);
    clear(nn,dd);
    getnext(x);getnext(x);
    ::read(x,o,n);
    for(int i=0;i<dim;i++)
      ::read(x,v[i],n);
  }

  void aff(ostream &cout=std::cout) const {
    if(dim<0) cout << "NULL\n";
    else {
      for(int i=0;i<n;i++)
	cout<<o[i] << " ";
      cout << endl;
      for(int d=0;d<dim;d++) {
	cout<< "  ";
	for(int i=0;i<n;i++)
	  cout<<v[d][i] << " ";
	cout << endl;
      }
    }
  }

  void affd(ostream &cout=std::cout) const {
    if(dim<0) cout << "NULL\n";
    else {
      for(int i=0;i<n;i++)
	cout<<getd(o[i]) << " ";
      cout << endl;
      for(int d=0;d<dim;d++) {
	cout<< "  ";
	for(int i=0;i<n;i++)
	  cout<<getd(v[d][i]) << " ";
	cout << endl;
      }
    }
  }
};

using ass_t = ass__t<rat_t>;

#endif

